"""
===============================================
Formal tests on the paths_generator, integrator
and SDE/SDEs classes
===============================================
"""
from .shared import *

process = sp.process
kfunc = sp.kfunc
paths_generator, integrator = sp.paths_generator, sp.integrator
SDE, SDEs = sp.SDE, sp.SDEs
integrate = sp.integrate

norm_rv = sp.norm_rv
uniform_rv = sp.uniform_rv
exp_rv = sp.exp_rv
double_exp_rv = sp.double_exp_rv

true_wiener_source = sp.true_wiener_source


# -----------------------------
# paths_generator general tests
# -----------------------------

# a generator subclass, to be tested
@kfunc
class generator_cls(paths_generator):

    depth = 2

    def __init__(self, *, exitf=1., vshape=(),
                 xw0=None, **args):
        self.exitf, self.xw0 = exitf, xw0
        self.vshape = vshape
        super().__init__(xshape=vshape, wshape=vshape,
                         **args)

    def begin(self):
        super().begin()
        iv = self.itervars
        iv['xw'][0][...] = np.asarray(self.xw0)

    def next(self):
        super().next()
        iv = self.itervars
        sw, xw = iv['sw'], iv['xw']
        x = xw[0]
        ds = sw[1] - sw[0]
        xw[1][...] = x + ds

    def store(self, i, k):
        super().store(i, k)
        iv = self.itervars
        iv['xx'][i] = iv['xw'][k]

    def exit(self, tt, xx):
        return process(tt, x=xx * self.exitf)


# tests with output timeline == integration timeline
def test_generator_t():
    np.random.seed(SEED)

    integr = generator_cls(paths=1, vshape=(), xw0=1)
    t = np.linspace(0, 10, 11)

    # test integration
    p = integr(t)
    assert_allclose(p[:, 0], 1 + t, rtol=eps(p.dtype))

    # test post processing with exit_integrate post-processing
    p = integr(t, exitf=10)
    assert_allclose(p[:, 0], 10 * (1 + t), rtol=eps(p.dtype))

    # test post processing with exit_integrate providing
    # explicit result
    p = integr(t, exitf=10)
    assert_allclose(p[:, 0], 10 * (1 + t), rtol=eps(p.dtype))

    # test multiple paths
    p = integr(t, paths=7)
    assert_allclose(p, np.full(t.shape + (7,), 1.) + t[:, np.newaxis],
                    rtol=eps(p.dtype))

    # test integration backwards
    p = integr(t)
    assert_allclose(p[:, 0], 1 + t, rtol=eps(p.dtype))

    # test exceptions
    assert_raises(ValueError, integr, t.reshape(-1, 1))  # bad shape
    assert_raises(ValueError, integr, (1, 2, 2, 3))  # repeated values
    assert_raises(ValueError, integr, (3, 2, 1))  # not sorted
    assert_raises(IndexError, integr, t, i0=Ellipsis)  # bad i0
    assert_raises(IndexError, integr, t, i0=12)  # bad i0
    assert_raises(IndexError, integr, t, i0=[1, 2, 3])  # bad i0
    integr2 = integr(vshape=())
    integr2.depth = 1
    assert_raises(ValueError, integr2, t)

    # test process dtype
    p = integr(t, dtype=np.float32)
    assert_(p.dtype == np.float32)
    assert_allclose(p[:, 0], 1 + t, rtol=eps(p.dtype))
    p = integr(t, dtype=np.float64)
    assert_(p.dtype == np.float64)
    assert_allclose(p[:, 0], 1 + t, rtol=eps(p.dtype))
    ti = np.arange(10)
    p = integr(ti, steps=37)
    assert_allclose(p[:, 0], 1 + ti, rtol=eps(p.dtype))


# tests with output timeline != integration timeline
def test_generator_steps():
    np.random.seed(SEED)

    integr = generator_cls(paths=1, vshape=(), xw0=1,
                           info={})
    t = np.linspace(0, 10, 11)

    # test integration and info
    p1 = integr(t)
    i1 = integr.info.copy()
    p2 = integr(t, steps=37)
    i2 = integr.info.copy()
    assert_allclose(p1, p2, rtol=eps(p1.dtype))
    assert_array_equal(p1.t, p2.t)
    assert_(i1['computed_steps'] == 10)
    assert_(i1['stored_steps'] == 10)
    assert_(i2['computed_steps'] >= 37)
    assert_(i2['stored_steps'] == 10)

    # test vshape same as wshape
    xw0 = np.arange(3*5*7).reshape(3, 5, 7)
    z = integr(paths=7, vshape=(3, 5), xw0=xw0)
    p1, p2 = z(t), z(t, steps=149)
    assert_allclose(p1, p2, rtol=16*eps(p1.dtype))


# testing depth > 2
class deep_cls(paths_generator):

    depth = 4

    def begin(self):
        super().begin()
        iv = self.itervars
        for k in range(3):
            iv['xw'][k][...] = (k + 1)*10

    def next(self):
        super().next()
        iv = self.itervars
        sw, xw = iv['sw'], iv['xw']
        x = xw[-2]
        ds = sw[-1] - sw[-2]
        xw[-1][...] = x + ds

    def store(self, i, k):
        super().store(i, k)
        iv = self.itervars
        iv['xx'][i] = iv['xw'][k]

    def exit(self, tt, xx):
        return process(tt, x=xx)


def test_generator_depth():

    np.random.seed(SEED)
    integr = deep_cls(dtype=int)
    x = integr((1, 2, 3))
    assert_((x == np.array((10, 20, 30)).reshape(-1, 1)).all())
    x = integr((1, 2, 3, 4, 5))
    assert_((x == np.array((10, 20, 30, 31, 32)).reshape(-1, 1)).all())
    assert_raises(ValueError, integr, (1, 2))
    assert_raises(ValueError, integr, (1,))


# ------------------------------
# integrator and integrate tests
# ------------------------------

def test_SDE():
    np.random.seed(SEED)

    paths = 3
    vshapes = ((), 2, (2, 5))
    t = (1., 2, 3, 4, 5, 6, 7)

    # 1-equation system
    # -----------------
    def f(t, x):
        return ({'dt': t+x, 'dw': x},)
    x = integrate(f)(x0=0.)(t)  # auto-detect q and sources
    assert_(len(x) == 1 and x[0].shape == (7, 1))

    # a minimal 2-equations system
    # ----------------------------
    def f(t, x, y=0.):
        return {'dt': 1, 'dw': 1}, {'dt': 1, 'dw': 1}

    # use defaults
    x, y = integrate(f)(x0=(1., 1.))(t)  # auto-detect
    for u in (x, y):
        assert_(u.shape == (7, 1))
    x, y = integrate(f, q=2, sources={'dt', 'dw'}  # explicit
                     )(x0=(1, 1))(t)
    for u in (x, y):
        assert_(u.shape == (7, 1))

    # evaluate with different shapes
    for vshape in vshapes:
        tst_vshape = (vshape,) if isinstance(vshape, int) else vshape
        x, y = integrate(f, q=2, sources={'dt', 'dw'})(
            x0=(1., 1.), paths=paths, vshape=vshape)(t)
        for z in (x, y):
            assert_(z.shape == (7,) + tst_vshape + (3,))

    # 3-equations system, parameters, different calling styles
    # --------------------------------------------------------
    def f(t, x=0, y=0, z=0, sigma=1., a=2.):
        return ({'dt': a*x+z, 'dw': sigma*y}, {'dt': t*y, 'dw': y+sigma*z},
                {'dt': a+x, 'dw': y-sigma*z})
    x, y, z = integrate(f)(x0=(1, 1, 1))(t)  # auto-detect
    for u in (x, y, z):
        assert_(u.shape == (7, 1))

    @integrate
    def f_process(t, x=0, y=0, z=0, sigma=1., a=2.):
        return f(t, x, y, z, sigma, a)

    x, y, z = f_process(sigma=2., x0=(1, 1, 1))(t)
    for u in (x, y, z):
        assert_(u.shape == (7, 1))

    @integrate(q=3, sources=('dt', 'dw'))
    def f_process(t, x=0, y=0, z=0, sigma=1., a=2.):
        return f(t, x, y, z, sigma, a)
    x, y, z = f_process(sigma=2., x0=(1, 1, 1))(t)
    for u in (x, y, z):
        assert_(u.shape == (7, 1))

    # same equation, calling with 'method' keyword
    # --------------------------------------------
    x, y, z = integrate(f)(method='euler', x0=(1, 1, 1))(t)
    for u in (x, y, z):
        assert_(u.shape == (7, 1))

    # same equation, time-dependent parameters, more parameters
    # ---------------------------------------------------------
    x, y, z = f_process(sigma=2., a=lambda t: 1+t,
                        vshape=5, paths=11, steps=30,
                        x0=(1, 1, 1))(t)
    for u in (x, y, z):
        assert_(u.shape == (7, 5, 11))

    # x0 and sigma with different values on the two elements of vshape
    x, y, z = integrate(f)(x0=(((1,), (1.1,)), ((2,), (2.1,)), ((3,), (3.1,))),
                           sigma=((.1,), (.2,)), a=lambda t: ((1+t,), (2*t,)),
                           vshape=2, paths=11, steps=30)(t)
    for u in (x, y, z):
        assert_(u.shape == (7, 2, 11))

    # path-dependent initial conditions, backward-forward integration
    xinit = np.linspace(1., 2., 11)
    x, y, z = f_process(sigma=((.1,), (.2,)),
                        x0=(xinit, 2*xinit, 3*xinit),
                        vshape=2, paths=11, steps=30, i0=5)(t)

    # same equation, time-dependent parameters,
    # with correlated wiener processes
    # -----------------------------------------
    args = dict(x0=(1, 2, 3), sigma=((.1,), (.2,)),
                a=lambda t: ((1+t,), (2*t,)),
                vshape=2, paths=11, steps=30)
    x, y, z = integrate(f)(**args, rho=lambda t: (.01*t, .2, .3))(t)
    for u in (x, y, z):
        assert_(u.shape == (7, 2, 11))
    x, y, z = integrate(f)(
        **args, corr=lambda t:
        ((1,   .01*t,  .2,  .3,  .4,  .5),
         (.01*t,   1, -.1, -.2, -.3, -.4),
         (.2,    -.1,   1,   0,  .1,   0),
         (.3,    -.2,   0,   1,  .1,  .2),
         (.4,    -.3,  .1,  .1,   1,   0),
         (.5,    -.4,   0,  .2,   0,   1))
        )(t)
    for u in (x, y, z):
        assert_(u.shape == (7, 2, 11))

    # 2 equations with poisson jumps
    # ------------------------------
    def f(t, x=0, y=0, k=1):
        return ({'dt': x, 'dn': k}, {'dt': 1, 'dn': k*y})
    x, y = integrate(f)(x0=(0., 0.), k=lambda t: .01*t,
                        lam=lambda t: .2*t)(t)
    for u in (x, y):
        assert_(u.shape == (7, 1))

    @integrate(q=2, sources={'dt', 'dn'})
    def f_process(t, x, y, k):
        return f(t, x, y, k)

    x, y = f_process(x0=(0., 0.), k=lambda t: .01*t,
                     lam=lambda t: .2*t, steps=30)(t)
    for u in (x, y):
        assert_(u.shape == (7, 1))

    # 2 equations with poisson jumps and wiener increments
    # ----------------------------------------------------
    def f(t, x=0, y=0, k=1):
        return ({'dt': x, 'dn': k, 'dw': y}, {'dt': 1, 'dn': k*y, 'dw': x-y})
    x, y = integrate(f)(x0=(0., 0.), k=lambda t: .01*t,
                        lam=lambda t: .2*t, rho=-.5)(t)
    for u in (x, y):
        assert_(u.shape == (7, 1))

    @integrate(q=2, sources={'dt', 'dn', 'dw'})
    def f_process(t, x, y, k):
        return f(t, x, y, k)

    x, y = f_process(x0=(0., 0.), k=lambda t: .01*t,
                     lam=lambda t: .2*t, rho=lambda t: -.01*t,
                     steps=30)(t)
    for u in (x, y):
        assert_(u.shape == (7, 1))

    # 2 equations with compound poisson jumps
    # ---------------------------------------
    def f(t, x=0, y=0, k=1):
        return ({'dt': x, 'dj': k}, {'dt': 1, 'dj': k*y})
    x, y = integrate(f)(x0=(0., 0.), k=lambda t: .01*t,
                        lam=lambda t: .2*t,
                        y=uniform_rv(a=-1, b=lambda t: t))(t)
    for u in (x, y):
        assert_(u.shape == (7, 1))

    @integrate(q=2, sources={'dt', 'dj'})
    def f_process(t, x, y, k):
        return f(t, x, y, k)

    x, y = f_process(x0=(0., 0.), k=lambda t: .01*t, lam=lambda t: .2*t,
                     y=exp_rv(a=-1), steps=30)(t)
    for u in (x, y):
        assert_(u.shape == (7, 1))

    # 2 equations with poisson jumps and wiener increments
    # ----------------------------------------------------
    def f(t, x=0, y=0, k=1):
        return ({'dt': x, 'dj': k, 'dw': y}, {'dt': 1, 'dj': k*y, 'dw': x-y})
    x, y = integrate(f)(x0=(0., 0.), k=lambda t: .01*t,
                        lam=lambda t: .2*t, rho=-.5,
                        y=norm_rv(a=1, b=lambda t: 2+t))(t)
    for u in (x, y):
        assert_(u.shape == (7, 1))

    @integrate(q=2, sources={'dt', 'dj', 'dw'})
    def f_process(t, x, y, k):
        return f(t, x, y, k)

    x, y = f_process(x0=(0., 0.), k=lambda t: .01*t,
                     lam=lambda t: .2*t, rho=lambda t: -.01*t,
                     y=double_exp_rv(a=1, b=2, pa=.1), steps=30)(t)
    for u in (x, y):
        assert_(u.shape == (7, 1))

    # minimal 4 equations with no 'dt' term
    # -------------------------------------
    def f(t, x=0, y=0, z=0, w=0):
        return ({'dw': x}, {'dw': y}, {'dw': z}, {'dw': w})
    rm4 = np.random.random((4, 4))
    corr4 = np.eye(4) + 0.1 * (rm4 + rm4.T)
    xs = integrate(f)(x0=(1,)*4, corr=corr4, paths=11, steps=30)(t)
    for u in xs:
        assert_(u.shape == (7, 11))

    # as above, with a true_wiener source
    # -----------------------------------
    tw = true_wiener_source(vshape=4, paths=11,
                            corr=corr4)
    xs1 = integrate(f)(x0=(1.,)*4, dw=tw, paths=11, steps=30)(t)
    for u in xs1:
        assert_(u.shape == (7, 11))
    xs2 = integrate(f)(x0=(1.,)*4, dw=tw, paths=11, steps=30)(t)
    for u, v in zip(xs1, xs2):
        assert_allclose(xs1, xs2, rtol=eps(u.dtype))



"""
========================
TESTS ON THE KFUNC CLASS
========================
"""
from .shared import *
kfunc = sp.kfunc
iskfunc = sp.iskfunc


def test_aaa_config():
    """testall.py: print realized configuration"""
    if VERBOSE:
        print('\n****', __name__, 'configuration:')
        print('KFUNC, iskfunc(wiener), iskfunc(wiener_process) =',
              KFUNC, sp.iskfunc(sp.wiener), sp.iskfunc(sp.wiener_process))
        print('PLOT, SAVE_ERRORS =',
              PLOT, SAVE_ERRORS)
        print('VERBOSE, QUANT_TEST_MODE =',
              VERBOSE, QUANT_TEST_MODE)


test_aaa_config.config = True


# -----------
# test kfunc
# -----------

def test_kfunc():

    # test kfunc definition errors
    # ----------------------------

    class fail1:
        def __new__(cls):
            pass

    class fail2:
        def __init__(self):
            pass

    class fail3:
        def __call__(self):
            pass

    class fail4(fail3):
        def __init__(self, x):
            pass

    class fail5:
        def __init__(self, *, x):
            pass

        def __call__(self, x):
            pass

    class warn1(fail2, fail3):
        params = 0

    @kfunc
    class warn2_parent(fail2, fail3):
        pass

    class warn2(warn2_parent):
        pass

    assert_raises(TypeError, kfunc, fail1)
    assert_raises(TypeError, kfunc, fail2)
    assert_raises(TypeError, kfunc, fail3)
    assert_raises(TypeError, kfunc, fail4)
    assert_raises(TypeError, kfunc, fail5)

    assert_warns(RuntimeWarning, kfunc, warn1)
    with assert_warns(RuntimeWarning):
        warn2()

    # define a test kfunc class
    # -------------------------
    @kfunc
    class Fbase:
        """
        kfunc with parmeters: a, b and info,
        variables: x, y
        info is used to return evaluation additional info
        """

        def __init__(self, *, a=1, b=2, info=None):
            # store in attributes what was passed to __init__
            self.a, self.b, self.info = a, b, info
            # initialize a info attribute
            if self.info is None:
                self.info = {}

        def __call__(self, x=11, y=22):
            self.info['value'] = (x, y)  # a test value returned by reference
            return (x, y, self.a, self.b)  # some test return values

    # define test subclasses of kfuncs
    # --------------------------------
    @kfunc
    class Finit(Fbase):

        def __init__(self, *, a=1, b=2, info=None):
            self.aa, self.bb = a, b
            super().__init__(a=a, b=b, info=info)

    @kfunc
    class Fcall(Fbase):

        def __call__(self, x=11, y=22):
            self.x, self.y = x, y
            return super().__call__(x, y)

    @kfunc
    class Fboth(Finit, Fcall):
        pass

    # tests an F instance against given expected args values
    # ------------------------------------------------------
    def F_check(f: 'F instance',
                aa: 'expected a', bb: 'expected b',
                ii: 'info' = None):

        # check args and attributes
        assert_(f.a == aa and f.b == bb)
        assert_(f.params == dict(a=aa, b=bb, info=ii))

        # check evaluation with default args
        assert_(f() == (11, 22, aa, bb))
        assert_(f.info['value'] == (11, 22))
        assert_(f('z', 'w') == ('z', 'w', aa, bb))
        assert_(f.info['value'] == ('z', 'w'))

        # check evaluation changing args
        # x=110, y is default, set a=3
        assert_(f(110, a=3) == (110, 22, 3, bb))
        assert_(f.a == aa and f.b == bb)  # defaults preserved
        if ii is not None:
            assert_(f.info is ii)  # info preserved and updated
            assert_(f.info['value'] == (110, 22))
        # x=110, set b=4
        assert_(f(1100, b=4) == (1100, 22, aa, 4))
        assert_(f.a == aa and f.b == bb)  # defaults preserved
        if ii is not None:
            assert_(f.info is ii)
            assert_(f.info['value'] == (1100, 22))
        # x=110, y=220, set a=3 and b=4, force info
        f_value = f(110, 220, a=3, b=4, info=f.info)  # set info dict
        assert_(f_value == (110, 220, 3, 4))
        assert_(f.a == aa and f.b == bb)  # defaults preserved
        assert_(f.info['value'] == (110, 220))  # info updated

    # do tests on F
    # -------------
    for F in (Fbase, Finit, Fcall, Fboth):

        # F instance with default args
        f = F()
        F_check(f, 1, 2)
        assert_(f._kfunc_parent is None)

        # F instance with non-default args
        f = F(a=11)
        F_check(f, 11, 2)
        f = F(b=13)
        F_check(f, 1, 13)
        f = F(a=11, b=13)
        F_check(f, 11, 13)

        # direct call of the class to instantiate and evaluate in one go
        assert_(F(110) == (110, 22, 1, 2))
        assert_(F('z', 'w') == ('z', 'w', 1, 2))
        assert_(F(110, a=3) == (110, 22, 3, 2))
        assert_(F(110, b=4) == (110, 22, 1, 4))
        assert_(F(110, a=3, b=4) == (110, 22, 3, 4))
        assert_(F('z', 'w', a=3, b=4) == ('z', 'w', 3, 4))

        # instantiaton from an instance, with modified args
        f = F(a=7)
        F_check(f, 7, 2)
        g = f(a=11, info=f.info)
        F_check(g, 11, 2, f.info)
        assert_(g._kfunc_parent is f)
        g = f(a=11, b=13, info=f.info)
        F_check(g, 11, 13, f.info)
        assert_(g._kfunc_parent is f)

        # type error in case of unexpected argument
        assert_raises(TypeError, F, z=1)  # at construction
        assert_raises(TypeError, F(), z=1, w=2)  # at instance from instance
        assert_raises(TypeError, F(), 11, 22, a=1, w=2)  # during evaluation
        assert_raises(TypeError, F, 11, 22, b=1, w=2)  # during evaluation

    # do tests on kfunc from functions
    # --------------------------------

    # define a kfunc from a function 1 variable and no parameters
    @kfunc(nvar=1)
    def G(x):
        return x

    assert_(G(5) == 5)
    assert_(G()(5) == 5)

    # define a kfunc from a function with 2 variables and parameters
    @kfunc(nvar=2)
    def G(x, y=0, *, p1='1', p2='2'):
        return x, y, p1, p2

    assert_(G(p1='x').params == dict(p1='x', p2='2'))
    assert_(G(1) == (1, 0, '1', '2'))
    assert_(G(1, 2) == (1, 2, '1', '2'))
    assert_(G(1, 2, p2='y') == (1, 2, '1', 'y'))
    assert_(G(1, p2='y') == (1, 0, '1', 'y'))
    assert_raises(TypeError, G(p3='z'), 1)

    g = G(p1='x', p2='y')
    assert_(g.params == dict(p1='x', p2='y'))
    assert_(g(1) == (1, 0, 'x', 'y'))
    assert_(g(1, p1='xx') == (1, 0, 'xx', 'y'))
    assert_(g(p1='xx')(1) == (1, 0, 'xx', 'y'))
    assert_(g(p1='xx')(1, 2) == (1, 2, 'xx', 'y'))
    assert_(g(p1='xx').params == dict(p1='xx', p2='y'))
    assert_raises(TypeError, g(p3='z'), 1)


def test_kfunc_exceptions():

    def fail():

        @kfunc(nvar=2)
        class A:
            pass

    assert_raises(SyntaxError, fail)

    def fail():

        @kfunc
        def f(x, y):
            pass

    assert_raises(SyntaxError, fail)

    def fail():

        @kfunc(nvar=3)  # nvar out of range
        def f(x, y):
            pass

    assert_raises(ValueError, fail)

    def fail():

        @kfunc(nvar=2)  # non-keyword parameter
        def f(x, y, z):
            pass

    assert_raises(TypeError, fail)


def test_kfunc_params():

    # sources
    # -------

    def kf(cls):
        if iskfunc(cls):
            return cls
        else:
            return kfunc(cls)

    for X in (sp.wiener_source, sp.dw):
        assert_(kf(X)(paths=2).params ==
                {'paths': 2, 'vshape': (), 'dtype': None,
                 'corr': None, 'rho': None})

    for X in (sp.odd_wiener_source, sp.odd_dw):
        # defaults beyond paths and vshape are missed
        # due to _antithetics wrapping
        assert_(kf(X)(paths=2).params ==
                {'paths': 2, 'vshape': ()})

    for X in (sp.true_wiener_source, sp.true_dw):
        assert_(kf(X)(paths=2).params ==
                {'paths': 2, 'vshape': (), 'dtype': None,
                 'rtol': 'max', 't0': 0., 'z0': 0.,
                'corr': None, 'rho': None})

    for X in (sp.poisson_source, sp.dn):
        assert_(kf(X)(lam=2, dtype=np.float32).params ==
                {'paths': 1, 'vshape': (),
                 'dtype': np.float32, 'lam': 2})

    # processes
    # ---------

    for X in (sp.wiener_process, sp.wiener):
        assert_(kf(X)(x0=2).params ==
                {'paths': 1, 'vshape': (), 'dtype': None,
                 'steps': None, 'i0': 0, 'info': None,
                 'getinfo': True, 'method': 'euler',
                 'dw': None, 'corr': None, 'rho': None,
                 'x0': 2, 'mu': 0.0, 'sigma': 1.0})

    for X in (sp.lognorm_process, sp.lognorm):
        assert_(kf(X)(x0=2).params ==
                {'paths': 1, 'vshape': (), 'dtype': None,
                 'steps': None, 'i0': 0, 'info': None,
                 'getinfo': True, 'method': 'euler',
                 'dw': None, 'corr': None, 'rho': None,
                 'x0': 2, 'mu': 0.0, 'sigma': 1.0})

    for X in (sp.ornstein_uhlenbeck_process, sp.oruh):
        assert_(kf(X)(theta=2, steps=20).params ==
                {'paths': 1, 'vshape': (), 'dtype': None,
                 'steps': 20, 'i0': 0, 'info': None,
                 'getinfo': True, 'method': 'euler', 'theta': 2,
                 'dw': None, 'corr': None, 'rho': None,
                 'x0': 0., 'k': 1., 'sigma': 1.})

    for X in (sp.hull_white_process, sp.hwff):
        assert_(kf(X)(vshape=3, factors=2).params ==
                {'paths': 1, 'vshape': 3, 'dtype': None,
                 'steps': None, 'i0': 0, 'info': None,
                 'getinfo': True, 'method': 'euler', 'theta': 0,
                 'dw': None, 'corr': None, 'rho': None,
                 'x0': 0., 'k': 1., 'sigma': 1.,
                 'factors': 2})

    for X in (sp.hull_white_1factor_process, sp.hw1f):
        assert_(kf(X)(vshape=3).params ==
                {'paths': 1, 'vshape': 3, 'dtype': None,
                 'steps': None, 'i0': 0, 'info': None,
                 'getinfo': True, 'method': 'euler', 'theta': 0.,
                 'dw': None, 'corr': None, 'rho': None,
                 'x0': 0., 'k': 1., 'sigma': 1.})

    for X in (sp.cox_ingersoll_ross_process, sp.cir):
        assert_(kf(X)(vshape=3).params ==
                {'paths': 1, 'vshape': 3, 'dtype': None,
                 'steps': None, 'i0': 0, 'info': None,
                 'getinfo': True, 'method': 'euler',
                 'dw': None, 'corr': None, 'rho': None,
                 'x0': 1., 'theta': 1., 'k': 1., 'xi': 1.})

    for X in (sp.heston_process, sp.heston,
              sp.full_heston_process, sp.heston_xy):
        assert_(kf(X)(vshape=2).params ==
                {'paths': 1, 'vshape': 2, 'dtype': None,
                 'steps': None, 'i0': 0, 'info': None,
                 'getinfo': True, 'method': 'euler',
                 'dw': None, 'corr': None, 'rho': None,
                 'x0': 1., 'y0': 1., 'mu': 0., 'sigma': 1.,
                 'theta': 1., 'k': 1., 'xi': 1.})

    for X in (sp.jumpdiff_process, sp.jumpdiff):
        assert_(kf(X)(vshape=2).params ==
                {'paths': 1, 'vshape': 2, 'dtype': None,
                 'steps': None, 'i0': 0, 'info': None,
                 'getinfo': True, 'method': 'euler', 'ptype': int,
                 'dw': None, 'corr': None, 'rho': None,
                 'dj': None, 'dn': None, 'lam': 1.,
                 'y': None, 'x0': 1., 'mu': 0., 'sigma': 1.})

    for X in (sp.merton_jumpdiff_process, sp.mjd):
        assert_(kf(X)(vshape=2).params ==
                {'paths': 1, 'vshape': 2, 'dtype': None,
                 'steps': None, 'i0': 0, 'info': None,
                 'getinfo': True, 'method': 'euler', 'ptype': int,
                 'dw': None, 'corr': None, 'rho': None,
                 'dj': None, 'dn': None, 'lam': 1.,
                 'a': 0., 'b': 1., 'x0': 1., 'mu': 0., 'sigma': 1.})

    for X in (sp.kou_jumpdiff_process, sp.kou):
        assert_(kf(X)(vshape=2).params ==
                {'paths': 1, 'vshape': 2, 'dtype': None,
                 'steps': None, 'i0': 0, 'info': None,
                 'getinfo': True, 'method': 'euler', 'ptype': int,
                 'dw': None, 'corr': None, 'rho': None,
                 'dj': None, 'dn': None, 'lam': 1.,
                 'a': 0.5, 'b': 0.5, 'pa': 0.5,
                 'x0': 1., 'mu': 0., 'sigma': 1.})



"""
This package provides tools to state and numerically
integrate Ito Stochastic Differential Equations (SDEs), including equations
with time-dependent parameters, time-dependent correlations, and
stochastic jumps, and to compute with, and extract statistics from,
their realized paths.

Package contents:

    1.  A set of tools to ease computations with stochastic processes,
        as obtained from numerical integration of the corresponding SDE,
        is provided via the ``process`` and ``montecarlo`` classes
        (see `Infrastructure`_):

        *   The ``process`` class, a subclass of ``numpy.ndarray`` representing
            a sequence of values in time, realized in one or several paths.
            Algebraic manipulations and ufunc computations are supported for
            instances that share the same timeline, or are constant, and
            comply with numpy broadcasting rules. Interpolation along
            the timeline is supported via callability of ``process`` instances.
            Process-specific functionalities, such as averaging and indexing
            along time or across paths, are delegated to process-specific
            methods, attributes and properties (no overriding
            of ``numpy.ndarray`` operations).

        *   The ``montecarlo`` class, as an aid to cumulate the results
            of several Monte Carlo simulations of a given
            stochastic variable, and to extract summary estimates
            for its probability distribution function and statistics.

    2.  Numerical realizations of the differentials commonly found
        as stochasticity sources in SDEs, are provided via
        the ``source`` class and its subclasses, with or without memory
        of formerly invoked realizations (see `Stochasticity Sources`_).

    3.  A general framework for stochastic step by step simulations,
        and for numerical SDE integration, is provided via the
        ``paths_generator`` class, and its cooperating subclasses
        ``integrator``, ``SDE`` and ``SDEs``
        (see `SDE Integration Framework`_).
        The full API allows for extensive customization of preprocessing,
        post-processing, stochasticity sources instantiation and handling,
        integration algorithms etc.
        The ``integrate`` decorator provides a simple and concise interface
        to handle standard use cases, via Euler-Maruyama integration.

    4.  Several preset stochastic processes are provided, including lognormal,
        Ornstein-Uhlenbeck, Hull-White n-factor, Heston, and
        jump-diffusion processes (see `Stochastic Processes`_).
        Each process consists of a process generator class,
        a subclass of ``integrator`` and ``SDE``, named with a
        ``_process`` suffix, and a definition of the underlying SDE,
        a subclass of ``SDE`` or ``SDEs``, named with a ``_SDE`` suffix.

    5.  Several analytical results relating to the preset stochastic
        processes are made available, as a general reference
        and for testing purposes (see `Analytical Results`_).
        They are limited to the case of constant process parameters,
        and with some further limitations on the parameters' domains.
        Function arguments are consistent with those of the
        corresponding processes. Suffixes ``_pdf``, ``_cdf`` and ``_chf``
        stand respectively for probability distribution
        function, cumulative probability distribution function,
        and characteristic function.
        Black-Scholes formulae for the valuation of call and put options
        have been included (with prefix ``bs``).

    6.  As an aid to interactive and notebook sessions, shortcuts are provided
        for stochasticity sources and preset processes (see `Shortcuts`_).
        Shortcuts have been wrapped as "kfuncs", objects with managed
        keyword arguments that simplify interactive workflow when
        frequent parameters tuning operations are needed
        (see ``kfunc`` decorator documentation).
        Analytical results are wrapped as kfuncs as well.

For all sources and processes, values can take any shape,
scalar or multidimensional. Correlated multivariate stochasticity sources are
supported. Poisson jumps are supported, and may be compounded with
any random variable supported by scipy.stats.
Time-varying process parameters (correlations, intensity of Poisson
processes, volatilities etc.) are allowed whenever applicable.
``process`` instances act as valid stochasticity source realizations (as does
any callable object complying with a ``source`` protocol), and may be
passed as a source specification when computing the realization of a given
process.

Computations are fully vectorized across paths, providing an efficient
infrastructure for simulating a large number of process realizations.
Less so, for large number of time steps: integrating 100 time steps
across one million paths takes seconds, one million time steps across
100 paths takes minutes.
"""

from .infrastructure import *
from .integration import *
from .analytical import *
from .kfun import *
from .shortcuts import *

import numpy.testing
test = numpy.testing.Tester().test

__version__ = '1.1.0'


_exclude = ('np', 'numpy', 'scipy',
            'exp', 'log', 'sqrt',
            'bisect', 'inspect', 'warnings',
            'infrastructure',
            'integration',
            'analytical',
            'kfun',
            'shortcuts',
            'tests',
            )

_include = ('__version__',)


__all__ = [s for s in dir() if
           not s.startswith('_') and s not in _exclude]

__all__ += _include



"""
===================================================================
FORMAL (FAST) AND QUANTITATIVE (SLOW) TESTS ON THE MONTECARLO CLASS
===================================================================
"""
from .shared import *

import scipy
import scipy.stats
import scipy.interpolate

montecarlo = sp.montecarlo


def test_aaa_config():
    """testall.py: print realized configuration"""
    if VERBOSE:
        print('\n****', __name__, 'configuration:')
        print('KFUNC, iskfunc(wiener), iskfunc(wiener_process) =',
              KFUNC, sp.iskfunc(sp.wiener), sp.iskfunc(sp.wiener_process))
        print('PLOT, SAVE_ERRORS =',
              PLOT, SAVE_ERRORS)
        print('VERBOSE, QUANT_TEST_MODE =',
              VERBOSE, QUANT_TEST_MODE)


test_aaa_config.config = True


# -------------
# test workflow
# -------------

# main test
def test_montecarlo_workflow():
    np.random.seed(SEED)

    # do cases
    shape = [(), (2,), (3, 2)]
    paths = [1, 10]
    dtype = [None, float, np.float64, np.float32,
             int, np.int64, np.int32]
    do(montecarlo_workflow, shape, paths, dtype)


# cases
def montecarlo_workflow(shape, paths, dtype):
    # print(shape, paths, dtype)
    PATHS = paths
    size = np.empty(shape).size
    indexes = [i for i in np.ndindex(shape)]

    # cumulate 10 simulations
    a = montecarlo(bins='auto')
    assert_raises(ValueError, a.histogram)
    for i in range(10):
        sample = np.random.normal(size=shape + (PATHS,)).astype(dtype)
        sample *= (1+np.arange(size)).reshape(shape + (1,))
        a.update(sample)
    assert_(a.paths == PATHS*10)

    # access results
    str(a)
    x = np.linspace(-3*size, 3*size, 100)
    for k in range(5):
        i = choice(indexes)
        a[i].pdf(x)  # default
        a[i].pdf(x, bandwidth=.7)  # gaussian kde with explicit bandwidth
        a[i].pdf(x, kind='linear', method='interp')
        a[i].pdf(x, kind='quadratic', method='interp')
        bins, counts = a[i].histogram()
        bins, counts = a[i].density_histogram()
        a[i].h, a[i].dh  # access using shortcuts
        a[i].outpaths, a[i].outerr()

    # access attributes
    assert_(a.paths == 10*PATHS)
    assert_(a.vshape == shape)
    assert_(a.shape == shape + (10*PATHS,))
    assert_array_equal(a.m, a.mean())
    assert_array_equal(a.m, a.stats['mean'])
    assert_array_equal(a.e, a.stderr())
    assert_array_equal(a.e, a.stats['stderr'])
    assert_allclose(a.e, a.stats['std']/np.sqrt(a.paths-1),
                    rtol=eps(a.m.dtype))

    # check dtype
    if len(shape) >= 1:
        type_ = np.dtype(dtype).type
        if type_(0.1) != type_(0):
            assert_(a.m.dtype == dtype)
            assert_(a.e.dtype == dtype)
        else:
            assert_(a.m.dtype == float)
            assert_(a.e.dtype == float)
    assert_(a._counts[i].dtype == a.ctype)

    # no histograms
    b = montecarlo(sample, bins=None)  # fist run
    b.stats['skew'], b.stats['kurtosis']
    assert_(b._bins is None)
    b.update(sample)  # cumulation run

    # explicit bins specification in native format
    c = montecarlo(bins=a._bins)
    c.update(sample)

    # explicit bins specification as array
    bins = np.empty(shape=shape + (10,))
    for k, i in enumerate(np.ndindex(shape), 1):
        bins[i] = np.linspace(-3*k, 3*k, 10)
    d = montecarlo(bins=bins)
    d.update(sample)

    # use paths along different axes
    e = None
    for k in range(len(shape)):
        axis_sample = np.random.normal(
            size=shape[:k] + (PATHS,) + shape[k:]).astype(dtype)
        # first run
        e = montecarlo(axis_sample, axis=k, bins=20, range=(-3, 3))
        # cumulation run
        e.update(axis_sample, axis=k)

    # non-array samples
    f = montecarlo(5)
    f.update(10)
    assert_(f.m == 7.5)
    f = montecarlo(((1, 2, 3), (2, 4, 6)), axis=1)
    f.update(((4, 5, 6), (8, 10, 12)), axis=1)

    assert_allclose(f.mean(), np.array((3.5, 7.)),
                    rtol=eps(f.mean().dtype))

    # antithetic sampling
    sample = np.random.normal(size=(2*PATHS,) + shape)
    even = montecarlo(sample, axis=0, use='even')
    odd = montecarlo(sample, axis=0, use='odd')
    for a in (even, odd):
        assert_(a.std().shape == shape)
        assert_(a.paths == PATHS)
    assert_allclose(even.mean(), montecarlo(sample, 0).mean())
    assert_allclose(odd.mean(), (sample[:PATHS]-sample[PATHS:]).mean(axis=0)/2)
    sample2 = np.random.normal(size=(4*PATHS,) + shape)
    even.update(sample2, axis=0)
    odd.update(sample2, axis=0)
    for a in (even, odd):
        assert_(a.mean().shape == shape)
        assert_(a.std().shape == shape)
        assert_(a.paths == 3*PATHS)
    even = montecarlo(use='even')
    odd = montecarlo(use='odd')
    even.update(sample, axis=0)
    odd.update(sample, axis=0)
    for a in (even, odd):
        assert_(a.mean().shape == shape)
        assert_(a.std().shape == shape)
        assert_(a.paths == PATHS)

    assert_raises(ValueError, montecarlo, sample, use=None)
    assert_raises(ValueError, montecarlo, np.zeros((3, 2)), axis=0,
                  use='even')


# -------------
# test cumulate
# -------------

# main test
def test_montecarlo_cumulate():
    np.random.seed(SEED)

    shape = [(), (2,), (3, 2)]
    paths = [1, 10]
    dtype = [None, float, np.float64, np.float32, np.float16,
             int, np.int64, np.int32, np.int16]
    do(montecarlo_cumulate, shape, paths, dtype)


# cases
def montecarlo_cumulate(shape, paths, dtype):
    PATHS = paths
    sample = np.random.normal(size=shape + (4*PATHS,)).astype(dtype)

    # full sample as first run
    a = montecarlo(sample)

    # same sample loaded in 3 runs
    b = montecarlo(sample[..., :PATHS])
    b.update(sample[..., PATHS:2*PATHS])
    b.update(sample[..., 2*PATHS:])

    rtol = max(eps(sample.dtype), eps(a.mean().dtype))
    assert_allclose(a.mean(), b.mean(), rtol=2*rtol)
    assert_allclose(a.stderr(), b.stderr(), rtol=rtol)
    assert_allclose(a.mean(), sample.mean(axis=-1),
                    rtol=rtol)
    assert_allclose(a.stderr(), sample.std(axis=-1)/np.sqrt(4*PATHS-1),
                    rtol=rtol)

    # same sample loaded in 3 runs, using same bins
    c = montecarlo()
    c._bins = a._bins
    c.update(sample[..., :PATHS])
    c.update(sample[..., PATHS:2*PATHS])
    c.update(sample[..., 2*PATHS:])
    for i in np.ndindex(c.vshape):
        assert_array_equal(a._counts[i], c._counts[i])
        assert_array_equal(a._bins[i], c._bins[i])
    assert_array_equal(a._paths_outside, c._paths_outside)


# -----------------------------
# test histogram (quantitative)
# context = 'histogram'
# -----------------------------

# main test

@slow
@quant
def test_montecarlo_histogram():
    np.random.seed(SEED)

    context = 'histogram'
    err_expected = load_errors(context)
    err_realized = {}

    BINS = 50
    PATHS = 1000*1000
    RATIO = 20  # ratio of last to first bin in nonlinear bins
    x = np.linspace(0, 1, BINS + 1)
    y = np.full(BINS + 1, RATIO**(1/BINS)).cumprod()
    y = (y - y.min())/(y.max() - y.min())
    assert_allclose((y.min(), y.max()), (0, 1))
    linear_bins = x
    nonlinear_bins = y

    #  dist_info: a dict of specs for testing montecarlo histograms,
    #  with the following structure:
    #    {test_id: (distribution instance, support, bins, paths)}
    dist_info = {
        'unif_lin': (scipy.stats.uniform(0, 1), (0, 1),
                     1.2*linear_bins - 0.1, PATHS),
        'unif_nonlin': (scipy.stats.uniform(0, 1), (0, 1),
                        1.2*nonlinear_bins - 0.1, PATHS),
        'norm_lin': (scipy.stats.norm(0, 1), (-3, 3),
                     8*linear_bins - 4, PATHS),
        'norm_nonlin': (scipy.stats.norm(0, 1), (-3, 3),
                        8*nonlinear_bins - 4, PATHS),
        'trapz_lin': (scipy.stats.trapz(0.2, 0.6), (0, 1),
                      1.2*linear_bins - 0.1, PATHS),
        'trapz_nonlin': (scipy.stats.trapz(0.2, 0.6), (0, 1),
                         1.2 * nonlinear_bins - 0.1, PATHS)
    }
    dist_keys = (
        'unif_lin', 'unif_nonlin', 'norm_lin', 'norm_nonlin',
        'trapz_lin', 'trapz_nonlin')  # make order predictable

    # fast test with several shapes and few paths, no error checking
    # and no plotting
    do(montecarlo_histogram, (context,), dist_keys,
       (noerrors_expected,), (noerrors_realized,),
       ((), (3,), (3, 5)),
       (dist_info,), (7,), (False,))

    # slow test with error checking and plotting
    do(montecarlo_histogram, (context,), dist_keys,
       (err_expected,), (err_realized,), ((3, 2),),
       (dist_info,), (PATHS,), (PLOT,))
    save_errors(context, err_realized)


# cases
def montecarlo_histogram(context, test_id, err_expected, err_realized,
                         shape, dist_info, paths, PLOT):
    dist, support, bins, _ = dist_info[test_id]
    test_key = context + '_' + test_id

    sample = dist.rvs(size=shape + (paths,))
    sample_bins = np.empty(shape + bins.shape)
    for i in np.ndindex(shape):
        sample_bins[i] = bins

    a = montecarlo(bins=sample_bins)
    p = paths//5
    a.update(sample[..., :p])
    a.update(sample[..., p:2*p])
    a.update(sample[..., 2*p:])

    mean_err = np.zeros(a.vshape)
    max_err = np.zeros(a.vshape)
    expected_counts = a.paths*np.diff(dist.cdf(bins))
    mean_counts = expected_counts[expected_counts > 0].mean()
    for i in np.ndindex(shape):
        counts, bins = a[i].histogram()
        delta = np.abs(counts - expected_counts)/mean_counts
        mean_err[i] = delta.mean()
        max_err[i] = delta.max()
        assert_quant(mean_err[i] < err_expected[test_key][0])
        assert_quant(max_err[i] < err_expected[test_key][1])

        ncounts, bins = a[i].density_histogram()
        assert_allclose((np.diff(bins)*ncounts).sum(), 1.,
                        rtol=eps(a.mean().dtype))

    err_realized[test_key] = (mean_err.max(), max_err.max())

    # plot worst case
    if PLOT:
        print('plotting...')
        i = (max_err == max_err.max())
        i = tuple(np.where(i)[k][0] for k in range(i.ndim))
        counts, bins = a[i].histogram()
        fig = plt.figure()
        plt.title(test_key + ':\nmean err = {:.5g}, max err = {:.5g}'
                  .format(float(mean_err[i]), float(max_err[i])) +
                  '\n error is conunts diff/mean counts')
        plot_histogram((expected_counts, bins), color='g',
                       histtype='step', label='expected', lw=2)
        plot_histogram((counts, bins), color='b',
                       histtype='step', label='realized')
        plt.plot(bins, bins*0 + mean_counts, 'c:', label='mean counts')
        plt.legend()
        plt.savefig(os.path.join(DIR, test_key + '.png'), dpi=300)
        plt.close(fig)


# -------------------------------
# test pdf and cdf (quantitative)
# context = 'distr'
# -------------------------------

# main test
@slow
@quant
def test_montecarlo_dist():
    np.random.seed(SEED)

    context = 'distr'
    err_expected = load_errors(context)
    err_realized = {}

    BINS = 50
    PATHS = 1000*1000
    RATIO = 10  # ratio of last to first bin in nonlinear bins
    x = np.linspace(0, 1, BINS + 1)
    y = np.full(BINS + 1, RATIO**(1/BINS)).cumprod()
    y = (y - y.min())/(y.max() - y.min())
    assert_allclose((y.min(), y.max()), (0, 1))
    linear_bins = x
    nonlinear_bins = y

    #  dist_info: a dict of specs for testing montecarlo histograms,
    #  with the following structure:
    #    {test_id: (distribution instance, support, bins, paths,
    #     integral_bounds)}
    dist_info = {
        'unif_lin': (scipy.stats.uniform(0, 1), (0, 1),
                     1.2*linear_bins - 0.1, PATHS, (-1, 2)),
        'unif_nonlin': (scipy.stats.uniform(0, 1), (0, 1),
                        1.2*nonlinear_bins - 0.1, PATHS, (-1, 2)),
        'norm_lin': (scipy.stats.norm(0, 1), (-3, 3),
                     8*linear_bins - 4, PATHS, (-6, 6)),
        'norm_nonlin': (scipy.stats.norm(0, 1), (-3, 3),
                        8*nonlinear_bins - 4, PATHS, (-6, 6)),
        'trapz_lin': (scipy.stats.trapz(0.2, 0.6), (0, 1),
                      1.2*linear_bins - 0.1, PATHS, (-1, 2)),
        'trapz_nonlin': (scipy.stats.trapz(0.2, 0.6), (0, 1),
                         1.2 * nonlinear_bins - 0.1, PATHS, (-1, 2))
    }
    dist_keys = (
        'unif_lin', 'unif_nonlin', 'norm_lin', 'norm_nonlin',
        'trapz_lin', 'trapz_nonlin')  # make order predictable

    # fast test with several shapes and few paths, no error checking
    # and no plotting
    do(montecarlo_dist, (context,), dist_keys,
       (noerrors_expected,), (noerrors_realized,),
       ((), (3,), (3, 5)),
       (dist_info,), (7,), (False,))

    # slow test with error checking and plotting
    do(montecarlo_dist, (context,), dist_keys,
       (err_expected,), (err_realized,), ((3, 2),),
       (dist_info,), (PATHS,), (PLOT,))
    save_errors(context, err_realized)


# cases
def montecarlo_dist(context, test_id, err_expected, err_realized,
                    shape, dist_info, paths, PLOT):
    dist, support, bins, _, integral_bounds = dist_info[test_id]
    test_key = context + '_' + test_id

    # prepare sample
    sample = dist.rvs(size=shape + (paths,))
    sample_bins = np.empty(shape + bins.shape)
    for i in np.ndindex(shape):
        sample_bins[i] = bins

    a = montecarlo(bins=sample_bins)
    p = paths//5
    a.update(sample[..., :p])
    a.update(sample[..., p:2*p])
    a.update(sample[..., 2*p:])

    # test pdf and cdf values
    (mean_err_pdf1, max_err_pdf1, mean_err_cdf1, max_err_cdf1,
     mean_err_pdf2, max_err_pdf2, mean_err_cdf2, max_err_cdf2) = \
        [np.zeros(a.vshape) for i in range(8)]
    x = np.linspace(bins[0], bins[-1], 10*bins.size)
    true_pdf = dist.pdf(x)
    true_cdf = dist.cdf(x)
    mean_pdf = 0.95/(4*dist.std())
    mean_cdf = 0.5
    for i in np.ndindex(shape):

        pdf1 = a[i].pdf(x, method='gaussian_kde')
        pdf2 = a[i].pdf(x, method='interp')
        delta1 = np.abs(pdf1 - true_pdf)/mean_pdf
        delta2 = np.abs(pdf2 - true_pdf)/mean_pdf
        mean_err_pdf1[i] = delta1.mean()
        mean_err_pdf2[i] = delta2.mean()
        max_err_pdf1[i] = delta1.max()
        max_err_pdf2[i] = delta2.max()
        assert_quant(mean_err_pdf1[i] <
                     err_expected[test_key + '_pdf_kde'][0])
        assert_quant(max_err_pdf1[i] <
                     err_expected[test_key + '_pdf_kde'][1])
        assert_quant(mean_err_pdf2[i] <
                     err_expected[test_key + '_pdf_interp'][0])
        assert_quant(max_err_pdf2[i] <
                     err_expected[test_key + '_pdf_interp'][1])

        cdf1 = a[i].cdf(x, method='gaussian_kde')
        cdf2 = a[i].cdf(x, method='interp')
        delta1 = np.abs(cdf1 - true_cdf)/mean_cdf
        delta2 = np.abs(cdf2 - true_cdf)/mean_cdf
        mean_err_cdf1[i] = delta1.mean()
        mean_err_cdf2[i] = delta2.mean()
        max_err_cdf1[i] = delta1.max()
        max_err_cdf2[i] = delta2.max()
        assert_quant(mean_err_cdf1[i] <
                     err_expected[test_key + '_cdf_kde'][0])
        assert_quant(max_err_cdf1[i] <
                     err_expected[test_key + '_cdf_kde'][1])
        assert_quant(mean_err_cdf2[i] <
                     err_expected[test_key + '_cdf_interp'][0])
        assert_quant(max_err_cdf2[i] <
                     err_expected[test_key + '_cdf_interp'][1])

    # store realized errors
    err_realized[test_key + '_pdf_kde'] = (mean_err_pdf1.max(),
                                           max_err_pdf1.max())
    err_realized[test_key + '_pdf_interp'] = (mean_err_pdf2.max(),
                                              max_err_pdf2.max())
    err_realized[test_key + '_cdf_kde'] = (mean_err_cdf1.max(),
                                           max_err_cdf1.max())
    err_realized[test_key + '_cdf_interp'] = (mean_err_cdf2.max(),
                                              max_err_cdf2.max())

    # test pdf normalization
    for i in np.ndindex(shape):
        y = np.linspace(integral_bounds[0], integral_bounds[1], 100*bins.size)
        counts, bins = a[i].histogram()
        integral1 = scipy.integrate.trapz(
            a[i].pdf(y, method='gaussian_kde'), y)
        integral2 = scipy.integrate.trapz(
            a[i].pdf(y, method='interp'), y)
        err1 = np.abs(integral1 - 1)
        err2 = np.abs(integral2 - 1)
        assert_quant(err1 < err_expected[test_key + '_integral_kde'][0])
        assert_quant(err2 < err_expected[test_key + '_integral_lin'][0])
        err_realized[test_key + '_integral_kde'] = (err1, err1)
        err_realized[test_key + '_integral_lin'] = (err2, err2)
        break  # test only first for speed

    # plot worst case
    if PLOT:
        print('plotting...')

        # plot pdf
        i = (max_err_pdf1 == max_err_pdf1.max())
        i = tuple(np.where(i)[k][0] for k in range(i.ndim))
        pdf1 = a[i].pdf(x, method='gaussian_kde')
        pdf2 = a[i].pdf(x, method='interp')
        counts, bins = a[i].density_histogram()
        fig = plt.figure()
        plt.title(test_key + '_pdf:\nmean err = {:.5g}, max err = {:.5g}'
                  .format(float(mean_err_pdf1[i]), float(max_err_pdf1[i])) +
                  '\n error is conunts diff/mean counts')
        plot_histogram((counts, bins), color='y',
                       histtype='step', label='realized histogram')
        plt.plot(x, true_pdf, color='g', label='expected pdf', lw=2)
        plt.plot(x, pdf1, color='b', label='realized pdf - gaussian kde')
        plt.plot(x, pdf2, color='r', label='realized pdf - interpolate')
        plt.plot(x, x*0 + mean_pdf, 'c:', label='mean pdf')
        plt.legend()
        plt.savefig(os.path.join(DIR, test_key + '_pdf.png'), dpi=300)
        plt.close(fig)

        # plot cdf
        i = (max_err_cdf1 == max_err_cdf1.max())
        i = tuple(np.where(i)[k][0] for k in range(i.ndim))
        cdf1 = a[i].cdf(x, method='gaussian_kde')
        cdf2 = a[i].cdf(x, method='interp')
        counts, bins = a[i].density_histogram()
        fig = plt.figure()
        plt.title(test_key + '_cdf:\nmean err = {:.5g}, max err = {:.5g}'
                  .format(float(mean_err_cdf1[i]), float(max_err_cdf1[i])) +
                  '\n error is conunts diff/mean counts')
        plot_histogram(
            hist=(
                (counts*np.diff(bins)).cumsum(),
                np.concatenate(((bins[:-1] + bins[1:])/2, bins[-1:]))
                ),
            color='y', histtype='step', label='realized histogram')
        plt.plot(x, true_cdf, color='g', label='expected cdf', lw=2)
        plt.plot(x, cdf1, color='b', label='realized cdf - gaussian kde')
        plt.plot(x, cdf2, color='r', label='realized cdf - interpolate')
        plt.legend()
        plt.savefig(os.path.join(DIR, test_key + '_cdf.png'), dpi=300)
        plt.close(fig)



"""
================================
INFRASTRUCTURE FOR FUNCTIONS AND
CLASSES WITH MANAGED KEYWORDS
================================

*  ``kfunc`` decorator,
*  ``iskfunc`` test.
"""

import inspect
import warnings

from .integration import SDE
# (replace with SDE = type(None)
# to make this module stand alone)


########################################
#  Private functions for recurring tasks
########################################

def _wraps(wrapped):
    """
    Decorator to preserve some basic attributes
    when wrapping a function or class.
    """
    def decorator(wrapper):
        for attr in ('__module__', '__name__', '__qualname__', '__doc__'):
            setattr(wrapper, attr, getattr(wrapped, attr))
        return wrapper
    return decorator


##########################################
#  kfunc decorator: function wrapper with
#  managed keyword arguments
##########################################

def _kfunc_split_args(x, args):
    call_args = {k: z for k, z in args.items()
                 if k in x._kfunc_call_args}
    init_args = {k: z for k, z in args.items()
                 if k not in x._kfunc_call_args}
    return call_args, init_args


class _meta(type):
    """
    Metaclass that calls only __new__, not __init__,
    upon instantiation.
    """
    def __call__(cls, *var, **args):
        return cls.__new__(cls, *var, **args)


def _kfunc_decorate_class(f):
    """
    Decorator to wrap the given class as a kfunc.
    """

    # try to prevent unexpected collateral damage
    # with some preemptive checks
    # -------------------------------------------

    def checkattr(attr):
        return any(attr in vars(base) for base in f.__mro__[:-1])

    if not hasattr(f, '_is_kfunc'):
        # if f is not already a kfunc subclass, make sure
        # that __new__ has not been messed with,
        # that __init__ and __call__ have been provided,
        # and warn against params attribute overwriting
        if checkattr('__new__'):
            raise TypeError('class {} has a a customized __new__ method, '
                            'should not be wrapped as a kfunc'
                            .format(f))
        if not (checkattr('__init__') and checkattr('__call__')):
            raise TypeError('wrapping {} as a kfunc, no user defined '
                            '__init__ and/or __call__ methods were found'
                            .format(f))
        if hasattr(f, 'params'):
            warnings.warn('wrapping {} as a kfunc, the params attribute '
                          'will be overwritten'.format(f),
                          RuntimeWarning)

    # get init and call signatures and enforce discipline:
    # - init arguments always passed as keywords
    # - no overlapping between call and init keywords
    init_signature = [(k, z) for k, z in
                      inspect.signature(f.__init__).parameters.items()
                      if not z.kind == z.VAR_KEYWORD][1:]  # discard self
    call_signature = [(k, z) for k, z in
                      inspect.signature(f.__call__).parameters.items()
                      if not z.kind == z.VAR_KEYWORD][1:]  # discard self

    if any(z.kind != z.KEYWORD_ONLY for k, z in init_signature):
        raise TypeError(
            'cannot wrap {} as a kfunc, all of its parameters '
            '(initialization arguments) should be keyword-only'
            .format(f))

    init_signature = dict([(k, z.default) for k, z in init_signature])
    call_signature = dict(call_signature)
    if not set(call_signature).isdisjoint(init_signature):
        raise TypeError(
            'cannot wrap {} as a kfunc, none of its parameters '
            '(initialization arguments) should be named as any '
            'of its variables (calling arguments)'
            .format(f))

    # create and return a wrapper class as a subclass of f
    # ----------------------------------------------------
    @_wraps(f)
    class kfunc_class_wrapper(f, metaclass=_meta):
        """
        Decorator to add kfunc functionality to a class.
        A wrapping subclass of f is returned.
        """

        _kfunc_init_args = init_signature
        _kfunc_call_args = call_signature

        def __new__(cls, *call_vars, **args):

            # avoid unsafe subclassing of kfunc classes
            if hasattr(cls, '_is_kfunc') and '_is_kfunc' not in vars(cls):
                warnings.warn(
                    'to prevent unexpected init and call behavior, '
                    'a subclass of a kfunc class should be decorated '
                    'with kfunc, but {} was not'.format(cls),
                    RuntimeWarning)
                self = f.__new__(cls)
                self.__init__(*call_vars, **args)
                return self

            # separate function arguments from instantiation parameters
            call_args, init_args = _kfunc_split_args(cls, args)

            # create a new class instance
            self = object.__new__(cls)
            self._kfunc_params = init_args
            self._kfunc_parent = None
            self.__init__(**init_args)

            # either return the instance, or call it and
            # return the result
            if call_vars or call_args:
                return super(cls, self).__call__(*call_vars, **call_args)
            else:
                return self
        __new__.__wrapped__ = f.__init__

        @_wraps(f.__call__)
        def __call__(self, *call_vars, **args):

            # separate function arguments from instantiation parameters
            call_args, init_args = _kfunc_split_args(self, args)

            if not init_args:
                # if no parameters, make a plain call to self
                return super().__call__(*call_vars, **call_args)
            else:
                # merge stored parameters with current ones
                # (current parameters override stored ones)
                new_init_args = {**self._kfunc_params, **init_args}

                # instantiate a derived object with merged parameters
                cls = type(self)
                new = object.__new__(cls)
                new.__init__(**new_init_args)
                new._kfunc_params = new_init_args
                new._kfunc_parent = self

                # handle evaluation/instantiation
                # with merged parameters
                if call_vars or call_args:
                    return super(cls, new).__call__(*call_vars, **call_args)
                else:
                    return new
        __call__.__wrapped__ = f.__call__

        if issubclass(f, SDE):
            # For integrators, the kfunc.params property is specialized
            # to include all parameters used by the SDE
            # (access is read-only)
            @property
            def params(self):
                return {**self._kfunc_init_args,
                        **self._kfunc_params,
                        **self.args}
        else:
            @property
            def params(self):
                return {**self._kfunc_init_args,
                        **self._kfunc_params}

        # mark the class as wrapped as a kfunc
        _is_kfunc = True

    return kfunc_class_wrapper


def _kfunc_decorate_function(nvar):
    """
    Decorator to wrap the given function as a kfunc.
    """

    def decorator(f):
        if isinstance(f, type):
            raise SyntaxError('improper use of kfunc decorator - see '
                              'kfunc docstring')

        f_signature = [(k, z) for k, z in
                       inspect.signature(f).parameters.items()
                       if not z.kind == z.VAR_KEYWORD]
        if not 0 < nvar <= len(f_signature):
            # avoid unexpected behaviour with nvar <= 0
            raise ValueError('expecting 0 < nvar <= {}, not {}'
                             .format(len(f_signature), nvar))
        if any(z.kind != z.KEYWORD_ONLY for k, z in f_signature[nvar:]):
            raise TypeError(
                'error wrapping {} as a kfunc - expecting nvar={} '
                'initial keyword or non-keyword arguments, '
                'as kfunc variables, followed by keyword arguments only, '
                'as kfunc parameters'
                .format(f, nvar))

        f_init_args = dict([(k, z.default) for k, z in f_signature[nvar:]])
        f_call_args = set(dict(f_signature[:nvar]))

        @_kfunc_decorate_class
        @_wraps(f)
        class kfunc_function_wrapper:
            def __init__(self, **args):
                self._kfunc_params = args

            def __call__(self, *var, **args):
                return f(*var, **args, **self._kfunc_params)
            __call__.__wrapped__ = f

        kfunc_function_wrapper._kfunc_init_args = f_init_args
        kfunc_function_wrapper._kfunc_call_args = f_call_args
        kfunc_function_wrapper.__wrapped__ = f

        return kfunc_function_wrapper

    return decorator


def kfunc(f=None, *, nvar=None):
    """
    Decorator to wrap classes or functions as objects
    with managed keyword arguments.

    This decorator, intended as an aid to interactive and notebook sessions,
    wraps a callable, class or function, as a "kfunc" object
    that handles separately its parameters (keyword-only),
    whose values are stored in the object, and its variables
    (positional or keyword), always provided upon evaluation.

    Syntax::

        @kfunc
        class my_class:
            def __init___(self, **kwparams):
                ...
            def __call__(self, *var, **kwvar):
                ...

        @kfunc(nvar=k)
        def my_function(*var, **kwargs):
            ...

    After decoration, ``my_class`` is a kfunc with ``kwparams`` as
    parameters, and with ``var`` and ``kwvar`` as variables, and
    ``my_function`` is a kfunc with the first ``k`` of ``var, kwargs``
    as variables, and the remaining ``kwargs`` as parameters.
    For usage, see examples below.

    Attributes
    ----------
    params : dictionary
       Parameter values stored in the instance (read-only). For
       wrapped ``SDE`` subclasses, also includes default values
       of all SDE-specific parameters, as stored in the ``args`` attribute.

    Examples
    --------
    Wrap ``wiener_source`` into a kfunc, named ``dw``:

        >>> from sdepy import wiener_source, kfunc
        >>> dw = kfunc(wiener_source)

    Instantiate ``dw`` and evaluate it (this is business as usual):

        >>> my_instance = dw(paths=100, dtype=np.float32)
        >>> x = my_instance(t=0, dt=1)
        >>> x.shape, x.dtype
        ((100,), dtype('float32'))

    Inspect kfunc parameters stored in ``my_instance``:

        >>> my_instance.params  # doctest: +SKIP
        {'paths': 100, 'vshape': (), 'dtype': <class 'numpy.float32'>, \
        'corr': None, 'rho': None}

    Evaluate ``my_instance`` changing some parameters (call
    the instance with one or more):

        >>> x = my_instance(t=0, dt=1, paths=999)
        >>> x.shape, x.dtype
        ((999,), dtype('float32'))

    Parameters stored in ``my_instance`` are not affected:

        >>> my_instance.paths == my_instance.params['paths'] == 100
        True

    Create a new instance, changing some parameters and keeping those
    already set (call the instance without passing any
    variables):

        >>> new_instance = my_instance(vshape=2, rho=0.5)
        >>> new_instance.params  # doctest: +SKIP
        {'paths': 100, 'vshape': 2, 'dtype': <class 'numpy.float32'>, \
        'corr': None, 'rho': 0.5}

    Instantiate and evaluate at once (pass one or more variables
    to the class constructor):

        >>> x = dw(0, 1, paths=100, dtype=np.float32)
        >>> x.shape, x.dtype
        ((100,), dtype('float32'))

    As long as variables are passed by name, order doesn't
    matter (omitted variables take default values, if any):

        >>> x = dw(paths=100, dtype=np.float32, dt=1, t=0)
        >>> x.shape, x.dtype
        ((100,), dtype('float32'))
    """
    # either f is a class and nvar is None,
    # or f is None and kfunc has been called with nvar
    # specified
    case1 = inspect.isclass(f) and nvar is None
    case2 = f is None and nvar is not None
    if not (case1 or case2):
        raise SyntaxError('improper use of kfunc decorator - see '
                          'kfunc docstring')
    if nvar is None:
        return _kfunc_decorate_class(f)
    else:
        return _kfunc_decorate_function(nvar)


def iskfunc(cls_or_object):
    """
    Tests if the given class or instance has been
    wrapped as a kfunc.
    """
    return hasattr(cls_or_object, '_is_kfunc')



"""
====================================
FORMAL TESTS ON PROCESS REALIZATIONS
====================================
"""
from .shared import *

from numpy import sqrt, exp
import scipy.stats

process = sp.process

wiener_source = sp.wiener_source
poisson_source = sp.poisson_source
cpoisson_source = sp.cpoisson_source

odd_wiener_source = sp.odd_wiener_source
even_poisson_source = sp.even_poisson_source
even_cpoisson_source = sp.even_cpoisson_source

true_wiener_source = sp.true_wiener_source
true_poisson_source = sp.true_poisson_source
true_cpoisson_source = sp.true_cpoisson_source

norm_rv = sp.norm_rv
uniform_rv = sp.uniform_rv
exp_rv = sp.exp_rv
double_exp_rv = sp.double_exp_rv
rvmap = sp.rvmap

integrator = sp.integrator
integrate = sp.integrate
iskfunc = sp.iskfunc

wiener_process = sp.wiener_process
lognorm_process = sp.lognorm_process
ornstein_uhlenbeck_process = sp.ornstein_uhlenbeck_process
hull_white_process = sp.hull_white_process
hull_white_1factor_process = sp.hull_white_1factor_process
cox_ingersoll_ross_process = sp.cox_ingersoll_ross_process
heston_process = sp.heston_process
full_heston_process = sp.full_heston_process
jumpdiff_process = sp.jumpdiff_process
merton_jumpdiff_process = sp.merton_jumpdiff_process
kou_jumpdiff_process = sp.kou_jumpdiff_process

wiener = sp.wiener
lognorm = sp.lognorm
oruh = sp.oruh
hwff = sp.hwff
hw1f = sp.hw1f
cir = sp.cir
heston = sp.heston
heston_xy = sp.heston_xy
jumpdiff = sp.jumpdiff
mjd = sp.mjd
kou = sp.kou


# -----------------------
# processes general tests
# -----------------------

all_cls = (wiener_process, lognorm_process,
           ornstein_uhlenbeck_process,
           hull_white_process, hull_white_1factor_process,
           cox_ingersoll_ross_process, full_heston_process, heston_process,
           jumpdiff_process, merton_jumpdiff_process,
           kou_jumpdiff_process)

all_shortcuts = (wiener, lognorm,
                 oruh,
                 hwff, hw1f,
                 cir, heston, heston_xy,
                 jumpdiff, mjd,
                 kou)


# enumerate test cases with constant parameters and launch tests
# --------------------------------------------------------------
def test_processes():
    np.random.seed(SEED)

    # setup some parameter test values
    # --------------------------------
    CM2 = ((1, .2), (.2, 1))
    CM3 = ((1, .2, -.3), (.2, 1, .1), (-.3, .1, 1))
    CM6 = np.eye(6) + 0.1*np.random.random((6, 6))
    CM6 = (CM6 + CM6.T)/2

    # processes to act as sources
    Z2 = wiener_process(sigma=.1, paths=11, vshape=2
                        )(np.linspace(0, 1, 100))
    Z23 = wiener_process(sigma=.1, paths=11, vshape=(2, 3)
                         )(np.linspace(0, 1, 100))
    Z32 = wiener_process(sigma=.1, paths=11, vshape=(3, 2)
                         )(np.linspace(0, 1, 100))

    # a 'bare' object with source protocol
    def S2(s, ds):
        return np.random.normal(size=(2, 11))*sqrt(np.abs(ds))

    S2.vshape = (2,)
    S2.paths = 11

    # a generic scipy random variable
    rv = scipy.stats.trapz(c=.1, d=.2)

    # test all classes with default parameters
    # and with generic dw
    # ----------------------------------------
    cls = all_cls + all_shortcuts
    paths, dtype = [1], [None]
    params = (dict(), dict(vshape=()), dict(vshape=(2,)))
    do(processes, cls, params, paths, dtype)

    paths, dtype = [3], [np.float16]
    params = (dict(),)
    do(processes, cls, params, paths, dtype)

    paths, dtype = [11], [None]
    params = (dict(vshape=2, dw=Z2),
              dict(vshape=2, dw=S2),
              )
    cls = list(cls)
    for p in (heston_process, heston,
              full_heston_process, heston_xy,
              hull_white_process, hwff):
        cls.remove(p)
    do(processes, cls, params, paths, dtype)

    # test each class with specific parameters
    # ----------------------------------------
    paths, dtype = [11], [None]

    cls = (wiener_process, lognorm_process,
           wiener, lognorm)
    params = (
        dict(vshape=2, x0=.1, mu=.2, sigma=.3),
        dict(vshape=2, x0=((.1,), (.11,)),
             mu=((.2,), (.21,)), sigma=((.3,), (.31,))),
        dict(vshape=(2,), corr=CM2),
        dict(vshape=(2, 3)),
        dict(vshape=(2, 3), corr=CM3),
        dict(vshape=(2, 3), dw=wiener_source(paths=11, vshape=(2, 3))),
        dict(vshape=(), dw=true_wiener_source(paths=11, vshape=())),
        dict(vshape=(2, 3), dw=Z23),
        )
    do(processes, cls, params, paths, dtype)

    cls = (ornstein_uhlenbeck_process, oruh)
    params = (
        dict(vshape=2, x0=.1, theta=.2, k=.4, sigma=.5),
        dict(vshape=(2,), corr=CM2),
        dict(vshape=(2,), dw=Z2),
        dict(vshape=(2, 3), dw=wiener_source(paths=11, vshape=(2, 3))),
        dict(vshape=(2, 3), dw=Z23),
        dict(vshape=(), dw=true_wiener_source(paths=11, vshape=()))
    )
    do(processes, cls, params, paths, dtype)

    cls = (hull_white_process, hwff)
    params = (
        dict(vshape=()),
        dict(vshape=(), factors=2),
        dict(vshape=(), factors=2,
             x0=((.1,), (.2,)), theta=((.4,), (.5,)), k=((.6,), (.7,)),
             sigma=((.8,), (.9,)), rho=0.25),
        dict(vshape=(), factors=2,
             dw=wiener_source(paths=11, vshape=(2,),
                              corr=((1, .25,), (.25, 1)))),
        dict(vshape=(3,), factors=2, dw=Z32),
        dict(vshape=(), factors=3,
             dw=wiener_source(paths=11, vshape=(3,), corr=CM3)),
        dict(vshape=(2, 3), factors=5,
             sigma=np.arange(2*3*5).reshape(2, 3, 5, 1)/(300)),
        dict(vshape=2, factors=3, dw=Z23),
        dict(vshape=(), factors=2, dw=Z2),
        dict(vshape=(), factors=2, dw=S2),
    )
    do(processes, cls, params, paths, dtype)

    cls = (hull_white_1factor_process, hw1f)
    params = (
        dict(vshape=2, x0=.1, theta=.2, k=.4, sigma=.5),
        dict(vshape=(2, 3), dw=wiener_source(paths=11, vshape=(2, 3))),
        dict(vshape=(), dw=true_wiener_source(paths=11, vshape=())),
    )
    do(processes, cls, params, paths, dtype)

    cls = (cox_ingersoll_ross_process, cir)
    params = (
        dict(vshape=2, x0=.1, theta=.2, k=.3, xi=.4),
        dict(vshape=(2,), corr=CM2),
        dict(vshape=(2,), rho=-.1),
        dict(vshape=(2, 3), dw=wiener_source(paths=11, vshape=(2, 3))),
        dict(vshape=(), dw=true_wiener_source(paths=11, vshape=())),
        dict(vshape=(2, 3), dw=Z23)
    )
    do(processes, cls, params, paths, dtype)

    cls = (full_heston_process, heston_xy)
    params = (
        dict(vshape=(1,)),
        dict(vshape=(3,), x0=.1, mu=.2, sigma=.3,
             y0=.4, theta=.5, k=.6, xi=.7),
        dict(vshape=(3,), x0=.1, mu=((.2,), (.3,), (.4,)), sigma=.3,
             y0=.4, theta=.5, k=.6, xi=.7),
        dict(vshape=(3,), corr=CM6),
        dict(vshape=(3,), rho=(.1, .2, .3)),
        dict(vshape=(5,), dw=wiener_source(paths=11, vshape=(10,))),
        dict(vshape=10, dw=true_wiener_source(paths=11, vshape=(20,))),
        dict(vshape=(), dw=Z2),
        dict(vshape=(), dw=S2),
    )
    do(processes, cls, params, paths, dtype)

    # correlation matrix should be even-dimensional and matching vshape
    assert_raises(ValueError, full_heston_process,
                  vshape=(2,), corr=CM6)
    assert_raises(ValueError, full_heston_process,
                  vshape=(2,), corr=np.asarray(CM6)[:5, :5])

    cls = (heston_process, heston)
    params = (
        dict(vshape=(1,)),
        dict(vshape=2, x0=.1, mu=.2, sigma=.3,
             y0=.4, theta=.5, k=.6, xi=.7),
        dict(vshape=(), corr=CM2),
        dict(vshape=(1,), corr=CM2),
        dict(vshape=(2, 3, 1), rho=.25),
        dict(vshape=(3), dw=wiener_source(paths=11, vshape=(6,))),
        dict(vshape=(2, 3, 1),
             dw=true_wiener_source(paths=11, vshape=(2, 3, 2))),
        dict(vshape=(), dw=Z2),
        dict(vshape=(), dw=S2),
    )
    do(processes, cls, params, paths, dtype)

    cls = (jumpdiff_process, jumpdiff,
           merton_jumpdiff_process, mjd,
           kou_jumpdiff_process, kou,
           )
    params = (
        dict(vshape=2, x0=.1, mu=.2, sigma=.3,
             lam=.4),
        dict(vshape=(2,), corr=CM2),
        dict(vshape=(2, 3), dw=wiener_source(paths=11, vshape=(2, 3))),
        dict(vshape=(3,), dw=true_wiener_source(paths=11, vshape=(3,)),
             dj=cpoisson_source(paths=11, vshape=(3,))),
        dict(vshape=(3,), dw=true_wiener_source(paths=11, vshape=(3,)),
             dj=true_cpoisson_source(paths=11, vshape=(3,))),
        dict(vshape=2, dj=poisson_source(paths=11, vshape=2)),
        dict(vshape=2, dj=true_poisson_source(paths=11, vshape=2)),
        dict(vshape=2, dj=cpoisson_source(paths=11, vshape=2, y=rv)),
        dict(vshape=2, dj=true_cpoisson_source(paths=11, vshape=2, y=rv)),
        dict(vshape=2, dw=Z2, dj=Z2),
        dict(vshape=2, dw=S2, dj=S2),
    )
    do(processes, cls, params, paths, dtype)

    cls = (jumpdiff_process, jumpdiff)
    params = (
        dict(vshape=2, y=norm_rv(a=.1, b=.2)),
        dict(vshape=2, y=uniform_rv(a=.1, b=.2)),
        dict(vshape=2, y=double_exp_rv(pa=.1, a=.2, b=.3)),
        dict(vshape=2, y=rv),
    )
    do(processes, cls, params, paths, dtype)

    cls = (merton_jumpdiff_process, mjd)
    params = (
        dict(vshape=2, x0=.1, mu=.2, sigma=.3,
             lam=.4, a=.5, b=.6),
    )
    do(processes, cls, params, paths, dtype)

    cls = (kou_jumpdiff_process, kou)
    params = (
        dict(vshape=()),
        dict(vshape=2, x0=.1, mu=.2, sigma=.3,
             lam=.4, pa=.5, a=.6, b=.7),
    )
    do(processes, cls, params, paths, dtype)


# enumerate test cases with time-dependent parameters and launch tests
# --------------------------------------------------------------------
def test_processes_local():
    np.random.seed(SEED)

    # setup some parameter test values
    # --------------------------------
    paths = [11]
    dtype = [None, float, np.float16]

    t = np.linspace(0, 2, 10)
    A = process(t, v=.1 + .2*t)
    B = A*A

    def C(t): return 0.1 + .1*t

    def D(t): return .1 + .5*t/10

    def CM2(t): return ((1, .2*t/10), (.2*t/10, 1))

    def CM6(t):
        C = np.eye(6) + t*0.1*np.random.random((6, 6))
        return (C + C.T)/2

    def R(t): return 0.5 - 0.1*t

    def R3(t): return (.1*C(t), .2*C(t), .3*C(t))

    CM2P = process((1., 2, 3), v=(CM2(1), CM2(2), CM2(3)))
    RP = A/10
    R3P = process((1., 2, 3), v=(R3(1), R3(2), R3(3)))

    # launch tests
    # --------------------------------
    cls = (wiener_process, lognorm_process)
    params = (
        dict(vshape=(), x0=.1, mu=A, sigma=B),
        dict(vshape=(2,), mu=A, sigma=C, corr=CM2),
        dict(vshape=(2,), mu=A, sigma=C, corr=CM2P),
        dict(vshape=(2,), mu=A, sigma=C, rho=R),
        dict(vshape=(2,), mu=A, sigma=C, rho=RP),
        dict(vshape=(6,), mu=A, sigma=C, corr=CM6),
        dict(vshape=(6,), mu=A, sigma=C, rho=R3P),
        dict(vshape=(6,), mu=A, sigma=C,
             dw=true_wiener_source(paths=11, vshape=6, rho=R3P)),
        )
    do(processes, cls, params, paths, dtype)

    cls = (ornstein_uhlenbeck_process, )
    params = (
        dict(vshape=(), x0=.1, theta=A, k=B, sigma=C),
        dict(vshape=(2,), theta=A, k=B, sigma=C, corr=CM2),
        dict(vshape=(2,), theta=A, k=B, sigma=C,
             dw=true_wiener_source(paths=11, vshape=2, corr=CM2)),
    )
    do(processes, cls, params, paths, dtype)

    cls = (hull_white_process, )
    params = (
        dict(vshape=(), factors=2, x0=((.1,), (.2,)),
             theta=lambda t: (A(t), B(t)), k=lambda t: ((C(t),), (D(t),)),
             sigma=lambda t: ((C(t),), (C(t),)), rho=C),
    )
    do(processes, cls, params, paths, dtype)

    cls = (cox_ingersoll_ross_process,)
    params = (
        dict(vshape=(), x0=.1, theta=A, k=B, xi=C),
        dict(vshape=(2,), theta=A, k=B, xi=C, corr=CM2)
    )
    do(processes, cls, params, paths, dtype)

    cls = (heston_process,)
    params = (
        dict(vshape=(), x0=.1, mu=A, sigma=B,
             y0=.2, theta=C, k=B, xi=A, rho=B),
        dict(vshape=(3,), theta=A, k=B, xi=C, corr=CM6),
        dict(vshape=(3,), theta=A, k=B, xi=C, rho=R3P)
    )
    do(processes, cls, params, paths, dtype)

    cls = (full_heston_process,)
    params = (
        dict(vshape=(), x0=.1, mu=A, sigma=B,
             y0=.2, theta=A, k=B, xi=C, rho=B),
        dict(vshape=(3,), theta=A, k=B, xi=C, corr=CM6)
    )
    do(processes, cls, params, paths, dtype)

    cls = (jumpdiff_process,)
    params = (
        dict(vshape=(), x0=.1, mu=A, sigma=B, lam=D),
        dict(vshape=(2,), mu=C, sigma=B, lam=A, corr=CM2),
        dict(vshape=(2,), mu=C, sigma=A, lam=B, corr=CM2P),
        dict(vshape=2, mu=A, sigma=B, y=norm_rv(a=D, b=C))
    )
    do(processes, cls, params, paths, dtype)

    cls = (merton_jumpdiff_process,)
    params = (
        dict(vshape=(), x0=.1, mu=A, sigma=B,
             lam=C, a=D, b=A),
        dict(vshape=2, mu=A, sigma=B, lam=C, corr=CM2),
    )
    do(processes, cls, params, paths, dtype)

    cls = (kou_jumpdiff_process,)
    params = (
        dict(vshape=(), x0=.1, mu=A, sigma=B,
             lam=C, pa=A, a=D, b=A),
        dict(vshape=(2,), mu=A, sigma=B, lam=C, corr=CM2)
    )
    do(processes, cls, params, paths, dtype)


# case testing
def processes(cls, params, paths, dtype):

    # print(cls.__name__, sep='', end='')
    P = cls(**params, paths=paths, dtype=dtype)

    vshape = params.get('vshape', P.vshape)
    if not isinstance(vshape, tuple):
        vshape = (vshape,)

    tlist = [(0.,), (1, 2), np.linspace(0, 1, 11)]
    stepslist = [None, 5, (1, 1.5, 2.5)]
    i0list = [0, 3, -2]
    for t in tlist:
        for steps in stepslist:
            for i0 in i0list:
                tt = np.asarray(t).reshape(-1)
                i0 = min(i0, tt.size - 1)
                if i0 + tt.size < 0:
                    i0 = 0
                if iskfunc(cls):
                    ps = P(t, i0=i0, steps=steps)
                else:
                    ps = cls(**params, paths=paths, dtype=dtype,
                             steps=steps, i0=i0)(t)
                if not isinstance(ps, (tuple, list)):
                    ps = (ps,)
                for p in ps:
                    assert_(isinstance(p, process))
                    assert_array_equal(p.t, tt)
                    assert_(p.vshape == vshape)
                    assert_(p.paths == paths)
                    assert_(p.dtype == dtype)
                    # no inf or nan
                    assert_(np.isfinite(p).all())
                    # initial values are the same for all paths
                    assert_((p[i0] == p[i0, ..., 0][..., np.newaxis]).all())

    # check initial conditions on the last computed p
    if cls in {wiener_process, lognorm_process,
               ornstein_uhlenbeck_process,
               cox_ingersoll_ross_process,
               jumpdiff_process,
               merton_jumpdiff_process,
               kou_jumpdiff_process}:
        assert_(len(ps) == 1)
        p = ps[0]
        p0, x0 = np.broadcast_arrays(p[i0], P.args['x0'])
        assert_allclose(p0, x0, rtol=eps(p.dtype))

    elif cls in {hull_white_process}:
        assert_(len(ps) == 1)
        p = ps[0]
        x0 = np.broadcast_to(P.args['x0'],
                             p.vshape + (P.args['factors'],) + (1,))
        p0, x0 = np.broadcast_arrays(p[i0], x0.sum(axis=-2))
        assert_allclose(p0, x0, rtol=eps(p.dtype))

    elif cls in {full_heston_process}:
        assert_(len(ps) == 2)
        p, q = ps
        p0, x0 = np.broadcast_arrays(p[i0], P.args['x0'])
        q0, y0 = np.broadcast_arrays(q[i0], P.args['y0'])
        assert_allclose(p0, x0, rtol=eps(p.dtype))
        assert_allclose(q0, y0, rtol=eps(p.dtype))


# stand alone test
def test_processes_exceptions():

    # timeline must be 1-dimensional
    for cls in all_cls:
        P = cls if iskfunc(cls) else cls()
        assert_raises(ValueError, P, np.arange(6.).reshape(2, 3))


def test_jumps():
    np.random.seed(SEED)

    # jump diffusion process integrated as is
    # (jumpdiff_process integrates the logarithm)
    @integrate(q=0, sources={'dt', 'dw', 'dj'})
    def jumpdiff_process2(s, x, mu=0., sigma=1.):
        return {'dt': mu*x,
                'dw': sigma*x,
                'dj': x}

    jdp1 = jumpdiff_process
    jdp2 = jumpdiff_process2
    t = (1, 1.5, 2)

    # compare using same stochasticity sources
    # (using almost deterministic y - easiest way to
    # synchronize dj1 and dj2 y values...)
    PATHS = 10
    eps = .000000001
    y1 = uniform_rv(.1, .1 + eps)
    y2 = rvmap(lambda y: exp(y) - 1, y1)
    dn = true_poisson_source(lam=10, paths=PATHS)
    dw = true_wiener_source(paths=PATHS)
    dj1 = true_cpoisson_source(dn=dn, paths=PATHS, y=y1)
    dj2 = true_cpoisson_source(dn=dn, paths=PATHS, y=y2)

    args = dict(sigma=.2, mu=0.2, paths=PATHS, steps=501)
    X1 = jdp1(**args, dj=dj1, dw=dw)
    X2 = jdp2(**args, dj=dj2, dw=dw)
    x1, x2 = X1(t), X2(t)
    # print(np.abs((x1-x2)/x1).max())
    # print(x1(2),x2(2))
    assert_(np.abs(1-x2/x1).max() < 0.02)

    # same as above using time dependence
    PATHS = 10
    eps = .000000001
    y1 = uniform_rv(lambda t: .05 + t*.05,
                    lambda t: .05 + t*.05 + eps)
    y2 = rvmap(lambda y: exp(y) - 1, y1)
    dn = true_poisson_source(lam=lambda t: t*5, paths=PATHS)
    dw = true_wiener_source(paths=PATHS)
    dj1 = true_cpoisson_source(dn=dn, paths=PATHS, y=y1)
    dj2 = true_cpoisson_source(dn=dn, paths=PATHS, y=y2)

    args = dict(sigma=lambda t: (.2 - t*.1), mu=lambda t: 0.1*t,
                paths=PATHS, steps=501)
    X1 = jdp1(**args, dj=dj1, dw=dw)
    X2 = jdp2(**args, dj=dj2, dw=dw)
    x1, x2 = X1(t), X2(t)
    # print(np.abs((x1-x2)/x1).max())
    # print(x1(2),x2(2))
    assert_(np.abs(1-x2/x1).max() < 0.03)

    # fully stochastic
    PATHS = 10000
    y1 = uniform_rv(0.05, 0.10)
    y2 = rvmap(lambda y: exp(y) - 1, y1)
    args = dict(sigma=.05, mu=-.1, lam=5, paths=PATHS, steps=501)
    X1 = jdp1(**args, y=y1)
    X2 = jdp2(**args, y=y2)
    x1, x2 = X1(t), X2(t)
    m1, s1 = x1[-1].mean(), x1[-1].std()
    m2, s2 = x2[-1].mean(), x2[-1].std()
    # print(m1, m2, s1, s2)
    assert_(abs(1-m2/m1) < 0.01)
    assert_(abs(1-s2/s1) < 0.03)


def test_processes_misc():

    # test exactness of wiener_process and lognorm_process
    # with constant parameters

    np.random.seed(SEED)
    paths = 31
    x0 = 10
    mu, sigma = .2, .7
    dw = true_wiener_source(paths=paths)
    t0, DT = 1, 3

    pw = wiener_process(x0=x0, mu=mu, sigma=sigma, paths=paths, dw=dw)
    pl = lognorm_process(x0=x0, mu=mu, sigma=sigma, paths=paths, dw=dw)

    t1 = (t0, t0 + DT/2, t0 + DT)
    t2 = np.linspace(t0, t0 + DT, 100)

    xw_exact = x0 + mu*DT + sigma*dw(t0, DT)
    xl_exact = x0*exp((mu - sigma*sigma/2)*DT + sigma*dw(t0, DT))
    xw1, xw2 = pw(t1), pw(t2)
    xl1, xl2 = pl(t1), pl(t2)

    rtol = eps(xw1.dtype)
    assert_allclose(xw_exact, xw1[-1], rtol=rtol)
    assert_allclose(xw_exact, xw2[-1], rtol=rtol)
    assert_allclose(xl_exact, xl1[-1], rtol=rtol)
    assert_allclose(xl_exact, xl2[-1], rtol=rtol)
